
/*
 *	Atom program to report details about Basic Blocks
 *	including:
 *		library/executable name
 *		Starting Address
 *		size of Bb (in number of instructions)
 *		redundancy of Bb's.
 *
 * Copyright (C) 1999   Rick Gorton gorton@world.std.com
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include <stdio.h>
#include <cmplrs/atom.inst.h>
#include "datatypes.h"
#include "declare.h"

typedef	struct {
	long	HCode;
	long	B_size;
	Block	*Bptr;
	void	*next;	/* Actually HBUCKET */
	long	dup_count;
	} HBUCKET;

HBUCKET	*HashTable[256];
long	DupBlockCount;
long	TotalBlocks;

void HashAddBlock(Block *this_block, int Block_len, long UseVec, long SetVec)
{
  int	index;
  long	Htmp;
  BOOLEAN	matched;
  HBUCKET	*walker;

  if (Block_len > 255)
    index = 0;
  else
    index = Block_len;

  matched = FALSE;
  Htmp = UseVec ^ SetVec;
  walker = HashTable[index];
  while ((walker != NULL) && !matched)
    {
    if ((Htmp == walker->HCode) && (Block_len == walker->B_size))
	{
	/*
	 * Walk both blocks to see if we get a match.
	 */
	Inst	*I_1;
	Inst	*I_2;
	int	i;
	long	I_1_addr, I_2_addr;
	int	I_1_bin, I_2_bin;

	I_1 = GetFirstInst(walker->Bptr);
	I_2 = GetFirstInst(this_block);
	for (i = 0; i < Block_len; i++)
	    {
	    I_1_bin = GetInstInfo(I_1, InstBinary);
	    I_2_bin = GetInstInfo(I_2, InstBinary);
	    if (I_1_bin != I_2_bin)
		{
		walker = walker->next;
		break;
		}
	    }
	if (i == Block_len)
	    {
	    walker->dup_count++;
	    DupBlockCount++;
	    matched = TRUE;
	    fprintf(Inst_Logfile, "Dup Blocks size %d :\t0x%lx :: 0x%lx\n",
				Block_len,
				InstPC(GetFirstInst(walker->Bptr)),
				InstPC(GetFirstInst(this_block)));
	    }
	}
    else
	walker = walker->next;
    }

  if (!matched)	/* Insert new item at front of chain */
    {
    HBUCKET	*Tptr;

    Tptr = (void *) malloc (sizeof(HBUCKET));
    Tptr->HCode = Htmp;
    Tptr->B_size = Block_len;
    Tptr->Bptr = this_block;
    Tptr->next = HashTable[index];
    Tptr->dup_count = 0x0;
    HashTable[index] = Tptr;
    }
}
void InstrumentInit(int iargc, char **iargv)
{
  int	i,j;
  char	namebuffer[256];

  if (iargc > 1)
    strcpy(namebuffer, iargv[1]);
  else
    strcpy(namebuffer, LOGFILE_NAME_);
  Inst_Logfile = fopen(namebuffer, "a+");
  printf("Logfile will be '%s'\n", namebuffer);

  TotalBlocks = 0x0;
  DupBlockCount = 0x0;
  for (i = 0; i < 255; i++)
    HashTable[i] = NULL;
  fprintf(Inst_Logfile, "Logfile:\t%s\n", namebuffer);
  fprintf(Inst_Logfile, "*** Bb data begin ***\n");
}

Instrument(int argc, char **argv, Obj *obj)
{
  Proc *p; Block *b;  Inst *this_inst;
  RegvType	RegA, RegB, RegC;
  long		this_PC;

  fprintf(Inst_Logfile, "Object:\t%s\n", GetObjName(obj));
  fprintf(Inst_Logfile, "Text Address:\t0x%lx\n", GetObjInfo(obj, ObjTextStartAddress));
  fprintf(Inst_Logfile, "Text Size:\t0x%lx\n", GetObjInfo(obj, ObjTextSize));
  fprintf(Inst_Logfile, "Data Address:\t0x%lx\n", GetObjInfo(obj, ObjInitDataStartAddress));
  fprintf(Inst_Logfile, "Data Size:\t0x%lx\n", GetObjInfo(obj, ObjInitDataSize));
  fprintf(Inst_Logfile, "BSS Address:\t0x%lx\n", GetObjInfo(obj, ObjUninitDataStartAddress));
  fprintf(Inst_Logfile, "BSS Size:\t0x%lx\n", GetObjInfo(obj, ObjUninitDataSize));
  fprintf(Inst_Logfile, "Number Procedures:\t0x%lx\n", GetObjInfo(obj, ObjNumberProcs));
  fprintf(Inst_Logfile, "Number Blocks:\t0x%lx\n", GetObjInfo(obj, ObjNumberBlocks));

  for (p = GetFirstObjProc(obj); p != NULL; p = GetNextProc(p))
    {
    fprintf(Inst_Logfile, "Procedure: <%s>\t@ 0x%lx\n", ProcName(p), ProcPC(p));
    for (b = GetFirstBlock(p); b != NULL; b = GetNextBlock(b))
	{
	InstRegUsageVec		this_bb_vec;
	long			block_len;

	this_bb_vec.ureg_bitvec[0] = 0x0;
	this_bb_vec.ureg_bitvec[1] = 0x0;

	this_bb_vec.dreg_bitvec[0] = 0x0;
	this_bb_vec.dreg_bitvec[1] = 0x0;

	block_len = GetBlockInfo(b, BlockNumberInsts);
	fprintf(Inst_Logfile, "Block @ 0x%lx\t", BlockPC(b));
	fprintf(Inst_Logfile, "length: 0x%lx\n", block_len);
	for (this_inst = GetFirstInst(b); this_inst != NULL;
			this_inst = GetNextInst(this_inst))
	    {
	    InstRegUsageVec		this_insn_vec;

	    GetInstRegUsage(this_inst, &this_insn_vec);

		/* Used */
	    this_bb_vec.ureg_bitvec[0] |= this_insn_vec.ureg_bitvec[0];
	    this_bb_vec.ureg_bitvec[1] |= this_insn_vec.ureg_bitvec[1];
		/* Set */
	    this_bb_vec.dreg_bitvec[0] |= this_insn_vec.dreg_bitvec[0];
	    this_bb_vec.dreg_bitvec[1] |= this_insn_vec.dreg_bitvec[1];
	    }	/* Each Inst in a block */

#ifdef WE_WANT_TO_SEE_USE_DEF_VECTORS
	fprintf(Inst_Logfile, "Use:\t0x%lx\t0x%lx\n",
				this_bb_vec.ureg_bitvec[0],
				this_bb_vec.ureg_bitvec[1]);
	fprintf(Inst_Logfile, "Set:\t0x%lx\t0x%lx\n",
				this_bb_vec.ureg_bitvec[0],
				this_bb_vec.ureg_bitvec[1]);
#endif
	TotalBlocks += block_len;
	if (block_len > 1)
	    {
		/*
		 * It is only worth looking at repeated block statistics
		 * for non-trivial blocks.
		 * Bb size data:
		 *	Renderman:	77787 Bbs, longest 0x1a6 insns
		 *	145.fpppp:	3146 Bbs, longest 0x170c insns
		 */
	    HashAddBlock(b, block_len, this_bb_vec.ureg_bitvec[0],
				this_bb_vec.dreg_bitvec[0]);
	    }
	}	/* Each block in a proc */
    }	/* Each proc */
}

void InstrumentFini()
{
  int	i;
  long	netval;
  HBUCKET	*walker;

  fprintf(Inst_Logfile, "*** Bb Summary begin ***\n");
  fprintf(Inst_Logfile, "Total blocks: %d\n", TotalBlocks);
  fprintf(Inst_Logfile, "Number of duplicated blocks: %d\n", DupBlockCount);
  fprintf(Inst_Logfile, "Breakout of duplicated blocks:\n");
  fprintf(Inst_Logfile, "\tSize\tRepeats\n");
  for (i = 2; i < 255; i++)
    {

    walker = HashTable[i];
    netval = 0x0;

    if (walker == NULL)
	continue;

    while (walker != NULL)
	{
	netval += walker->dup_count;
	walker = walker->next;
	}
    if (netval > 0)
	fprintf(Inst_Logfile, "\t%d\t%d\n", i, netval);
    }
  walker = HashTable[0];	/* Special - slot zero holds _big_ blocks */
  while (walker != NULL)
    {
    if (walker->dup_count > 0)
	fprintf(Inst_Logfile, "\t%d\t%d\n", walker->B_size, walker->dup_count);
    walker = walker->next;
    }

  fprintf(Inst_Logfile, "*** Bb data end ***\n");
  fclose(Inst_Logfile);

}

