
/*
 * Types/enumerations for [Atom] DBO (Dynamic binary optimization)
 * investigation programs.
 *
 * Copyright (C) 1999   Rick Gorton gorton@world.std.com
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include <machine/inst.h>

#ifndef TRUE
#define TRUE 1
#endif

#ifndef FALSE
#define FALSE 0
#endif

typedef unsigned char	U8;
typedef unsigned char	u8;
typedef u8	BOOLEAN;

	/* Shifts */
#define	SLL_LIT_OPCODE		0x48001720
#define	SLL_REG_OPCODE		0x48000720

#define	SRL_LIT_OPCODE		0x48001680
#define	SRL_REG_OPCODE		0x48000680

#define	SRA_LIT_OPCODE		0x48001780
#define	SRA_REG_OPCODE		0x48000780

	/* Logicals */
#define	AND_LIT_OPCODE 		0x44001000
#define	AND_REG_OPCODE 		0x44000000

#define	BIC_LIT_OPCODE 		0x44001100
#define	BIC_REG_OPCODE 		0x44000100

#define	BIS_LIT_OPCODE 		0x44001400
#define	BIS_REG_OPCODE 		0x44000400

#define	EQV_LIT_OPCODE 		0x44001900
#define	EQV_REG_OPCODE 		0x44000900

#define	ORNOT_LIT_OPCODE 	0x44001500
#define	ORNOT_REG_OPCODE 	0x44000500

#define	XOR_LIT_OPCODE 		0x44001800
#define	XOR_REG_OPCODE 		0x44000800

	/* Extracts */
#define	EXTBL_LIT_OPCODE	0x480010c0
#define	EXTBL_REG_OPCODE	0x480000c0

#define	EXTWL_LIT_OPCODE	0x480012c0
#define	EXTWL_REG_OPCODE	0x480002c0

#define	EXTLL_LIT_OPCODE	0x480014c0
#define	EXTLL_REG_OPCODE	0x480004c0

#define	EXTQL_LIT_OPCODE	0x480016c0
#define	EXTQL_REG_OPCODE	0x480006c0

#define	EXTWH_LIT_OPCODE	0x48001b40
#define	EXTWH_REG_OPCODE	0x48000b40

#define	EXTLH_LIT_OPCODE	0x48001d40
#define	EXTLH_REG_OPCODE	0x48000d40

#define	EXTQH_LIT_OPCODE	0x48001f40
#define	EXTQH_REG_OPCODE	0x48000f40

	/* Inserts */
#define	INSBL_LIT_OPCODE	0x48001160
#define	INSBL_REG_OPCODE	0x48000160

#define	INSWL_LIT_OPCODE	0x48001360
#define	INSWL_REG_OPCODE	0x48000360

#define	INSLL_LIT_OPCODE	0x48001560
#define	INSLL_REG_OPCODE	0x48000560

#define	INSQL_LIT_OPCODE	0x48001760
#define	INSQL_REG_OPCODE	0x48000760

#define	INSWH_LIT_OPCODE	0x48001ae0
#define	INSWH_REG_OPCODE	0x48000ae0

#define	INSLH_LIT_OPCODE	0x48001ce0
#define	INSLH_REG_OPCODE	0x48000ce0

#define	INSQH_LIT_OPCODE	0x48001ee0
#define	INSQH_REG_OPCODE	0x48000ee0

	/* Masks */
#define	MSKBL_LIT_OPCODE	0x48001040
#define	MSKBL_REG_OPCODE	0x48000040

#define	MSKWL_LIT_OPCODE	0x48001240
#define	MSKWL_REG_OPCODE	0x48000240

#define	MSKLL_LIT_OPCODE	0x48001440
#define	MSKLL_REG_OPCODE	0x48000440

#define	MSKQL_LIT_OPCODE	0x48001640
#define	MSKQL_REG_OPCODE	0x48000640

#define	MSKWH_LIT_OPCODE	0x48001a40
#define	MSKWH_REG_OPCODE	0x48000a40

#define	MSKLH_LIT_OPCODE	0x48001c40
#define	MSKLH_REG_OPCODE	0x48000c40

#define	MSKQH_LIT_OPCODE	0x48001e40
#define	MSKQH_REG_OPCODE	0x48000e40

	/* Zappage */
#define	ZAP_LIT_OPCODE		0x48001600
#define	ZAP_REG_OPCODE		0x48000600
#define	ZAPNOT_LIT_OPCODE	0x48001620
#define	ZAPNOT_REG_OPCODE	0x48000620

	/* Misc */
#define	AMASK_LIT_OPCODE	0x44001c20
#define	AMASK_REG_OPCODE	0x44000c20

#define	IMPLVER_OPCODE		0x44000d80

	/* Extends */
#define	SEXTB_OPCODE		0x70000000
#define	SEXTW_OPCODE		0x70000020

	/* Load */
#define	LDA_OPCODE		0x20000000
#define	LDAH_OPCODE		0x24000000
#define	LDBU_OPCODE		0x28000000
#define	LDWU_OPCODE		0x30000000
#define	LDL_OPCODE		0xa0000000
#define	LDQ_OPCODE		0xa4000000
#define	LDQU_OPCODE		0x2c000000

	/* Store */
#define	STB_OPCODE		0x38000000
#define	STW_OPCODE		0x34000000
#define	STL_OPCODE		0xb0000000
#define	STQ_OPCODE		0xb4000000
#define	STQU_OPCODE		0x3c000000

#define	OPCODE_END		0xffff
#define	FUNCTION_CODE_END	0xffff
#define	MAX_OPCODES		0xff


enum Opcodes {
	SLL_LIT_OP,	SLL_REG_OP,
	SRL_LIT_OP,	SRL_REG_OP,
	SRA_LIT_OP,	SRA_REG_OP,

	AND_LIT_OP,	AND_REG_OP,
	BIC_LIT_OP,	BIC_REG_OP,
	BIS_LIT_OP,	BIS_REG_OP,
	EQV_LIT_OP,	EQV_REG_OP,
	ORNOT_LIT_OP,	ORNOT_REG_OP,
	XOR_LIT_OP,	XOR_REG_OP,

	EXTBL_LIT_OP,	EXTBL_REG_OP,
	EXTWL_LIT_OP,	EXTWL_REG_OP,
	EXTLL_LIT_OP,	EXTLL_REG_OP,
	EXTQL_LIT_OP,	EXTQL_REG_OP,
	EXTWH_LIT_OP,	EXTWH_REG_OP,
	EXTLH_LIT_OP,	EXTLH_REG_OP,
	EXTQH_LIT_OP,	EXTQH_REG_OP,

	INSBL_LIT_OP,	INSBL_REG_OP,
	INSWL_LIT_OP,	INSWL_REG_OP,
	INSLL_LIT_OP,	INSLL_REG_OP,
	INSQL_LIT_OP,	INSQL_REG_OP,
	INSWH_LIT_OP,	INSWH_REG_OP,
	INSLH_LIT_OP,	INSLH_REG_OP,
	INSQH_LIT_OP,	INSQH_REG_OP,

	MSKBL_LIT_OP,	MSKBL_REG_OP,
	MSKWL_LIT_OP,	MSKWL_REG_OP,
	MSKLL_LIT_OP,	MSKLL_REG_OP,
	MSKQL_LIT_OP,	MSKQL_REG_OP,
	MSKWH_LIT_OP,	MSKWH_REG_OP,
	MSKLH_LIT_OP,	MSKLH_REG_OP,
	MSKQH_LIT_OP,	MSKQH_REG_OP,

	ZAP_LIT_OP,	ZAP_REG_OP,
	ZAPNOT_LIT_OP,	ZAPNOT_REG_OP,

	AMASK_LIT_OP,	AMASK_REG_OP,

	IMPLVER_OP,

	SEXTB_OP,
	SEXTW_OP,

	LDA_OP,
	LDAH_OP,
	LDBU_OP,
	LDWU_OP,
	LDL_OP,
	LDQ_OP,
	LDQU_OP,

	STB_OP,
	STW_OP,
	STL_OP,
	STQ_OP,
	STQU_OP,

	LAST_OP
	};

static const int Opcode_Value[(int) MAX_OPCODES] = {
	SLL_LIT_OPCODE,	SLL_REG_OPCODE,
	SRL_LIT_OPCODE,	SRL_REG_OPCODE,
	SRA_LIT_OPCODE,	SRA_REG_OPCODE,

	AND_LIT_OPCODE,	AND_REG_OPCODE,
	BIC_LIT_OPCODE,	BIC_REG_OPCODE,
	BIS_LIT_OPCODE,	BIS_REG_OPCODE,
	EQV_LIT_OPCODE,	EQV_REG_OPCODE,
	ORNOT_LIT_OPCODE,	ORNOT_REG_OPCODE,
	XOR_LIT_OPCODE,	XOR_REG_OPCODE,

	EXTBL_LIT_OPCODE,	EXTBL_REG_OPCODE,
	EXTWL_LIT_OPCODE,	EXTWL_REG_OPCODE,
	EXTLL_LIT_OPCODE,	EXTLL_REG_OPCODE,
	EXTQL_LIT_OPCODE,	EXTQL_REG_OPCODE,
	EXTWH_LIT_OPCODE,	EXTWH_REG_OPCODE,
	EXTLH_LIT_OPCODE,	EXTLH_REG_OPCODE,
	EXTQH_LIT_OPCODE,	EXTQH_REG_OPCODE,

	INSBL_LIT_OPCODE,	INSBL_REG_OPCODE,
	INSWL_LIT_OPCODE,	INSWL_REG_OPCODE,
	INSLL_LIT_OPCODE,	INSLL_REG_OPCODE,
	INSQL_LIT_OPCODE,	INSQL_REG_OPCODE,
	INSWH_LIT_OPCODE,	INSWH_REG_OPCODE,
	INSLH_LIT_OPCODE,	INSLH_REG_OPCODE,
	INSQH_LIT_OPCODE,	INSQH_REG_OPCODE,

	MSKBL_LIT_OPCODE,	MSKBL_REG_OPCODE,
	MSKWL_LIT_OPCODE,	MSKWL_REG_OPCODE,
	MSKLL_LIT_OPCODE,	MSKLL_REG_OPCODE,
	MSKQL_LIT_OPCODE,	MSKQL_REG_OPCODE,
	MSKWH_LIT_OPCODE,	MSKWH_REG_OPCODE,
	MSKLH_LIT_OPCODE,	MSKLH_REG_OPCODE,
	MSKQH_LIT_OPCODE,	MSKQH_REG_OPCODE,

	ZAP_LIT_OPCODE,		ZAP_REG_OPCODE,
	ZAPNOT_LIT_OPCODE,	ZAPNOT_REG_OPCODE,

	AMASK_LIT_OPCODE,	AMASK_REG_OPCODE,

	IMPLVER_OPCODE,

	SEXTB_OPCODE,
	SEXTW_OPCODE,

	LDA_OPCODE,
	LDAH_OPCODE,
	LDBU_OPCODE,
	LDWU_OPCODE,
	LDL_OPCODE,
	LDQ_OPCODE,
	LDQU_OPCODE,

	STB_OPCODE,
	STW_OPCODE,
	STL_OPCODE,
	STQ_OPCODE,
	STQU_OPCODE,

	OPCODE_END
	};

static const char Opcode_String[(int) MAX_OPCODES][32] = {
	/*
	 * Usage note:
	 *	_L suffix stands for 'literal'
	 *	_R suffix stands for 'register'
	 */
	"SLL_L",	"SLL_R",
	"SRL_L",	"SRL_R",
	"SRA_L",	"SRA_R",

	"AND_L",	"AND_R",
	"BIC_L",	"BIC_R",
	"BIS_L",	"BIS_R",
	"EQV_L",	"EQV_R",
	"ORNOT_L",	"ORNOT_R",
	"XOR_L",	"XOR_R",

	"EXTBL_L",	"EXTBL_R",
	"EXTWL_L",	"EXTWL_R",
	"EXTLL_L",	"EXTLL_R",
	"EXTQL_L",	"EXTQL_R",
	"EXTWH_L",	"EXTWH_R",
	"EXTLH_L",	"EXTLH_R",
	"EXTQH_L",	"EXTQH_R",

	"INSBL_L",	"INSBL_R",
	"INSWL_L",	"INSWL_R",
	"INSLL_L",	"INSLL_R",
	"INSQL_L",	"INSQL_R",
	"INSWH_L",	"INSWH_R",
	"INSLH_L",	"INSLH_R",
	"INSQH_L",	"INSQH_R",

	"MSKBL_L",	"MSKBL_R",
	"MSKWL_L",	"MSKWL_R",
	"MSKLL_L",	"MSKLL_R",
	"MSKQL_L",	"MSKQL_R",
	"MSKWH_L",	"MSKWH_R",
	"MSKLH_L",	"MSKLH_R",
	"MSKQH_L",	"MSKQH_R",

	"ZAP_L",	"ZAP_R",
	"ZAPNOT_L",	"ZAPNOT_R",

	"AMASK_L",	"AMASK_R",

	"IMPLVER_R",

	"SEXTB",
	"SEXTW",

	"LDA",
	"LDAH",
	"LDBU",
	"LDWU",
	"LDL",
	"LDQ",
	"LDQU",

	"STB",
	"STW",
	"STL",
	"STQ",
	"STQU",

	"CONSISTENCY ERROR"
	};

typedef union {
		long	longstuff;
		double	doublestuff;
		struct {
			long	mantissa:52;
			long	exponent:11;
			long	sign:1;
			} rawbits;
		} IEEE_UNION_;

typedef struct {
	long	counter;
	} INSN_TYPE;
/*
 * Structures for Atom-izing time data collection
 */
static long static_count[(int)LAST_OP];

/*
 * Structures for execution time data collection
 */
long dyn_count[(int)LAST_OP];
long total_dyn_count;

FILE	*Inst_Logfile;

/*
 * Macros (YUK) and constants for extracting the LIT field from an
 * operate instruction, since the ATOM people didn't deem it important
 * enough to provide this feature.  (Grumble)
 */

#define	OPCODE_MASK		0xfc000000
#define	OPERATE_MASK		0xfc001fe0
#define LITERAL_REGISTER_MASK	0x1000
#define LITERAL_FIELD_MASK	0x1fe000
#define LITERAL_SHIFT_COUNT	13
#define LITERAL_BIT_MASK	0xff

#define IS_LITERAL_OPERATE_(OPCODE32)	\
	(OPCODE32 & LITERAL_REGISTER_MASK)
#define LITERAL_VALUE_(OPCODE32)	\
	((OPCODE32 >> LITERAL_SHIFT_COUNT) & LITERAL_BIT_MASK)

